import {LitElement, html, unsafeHTML} from './lit-all.min.js';
import * as tfrpc from '/static/tfrpc.js';


class BlogAppElement extends LitElement {
	static get properties() {
		return {
			ids: {type: Array},
			id: {type: String},
			posts: {type: Array},
			post: {type: Object},
			published: {type: Array},
		};
	}

	constructor() {
		super();
		let self = this;
		this.ids = [];
		this.posts = [];
		this.published = [];
		this.load();
	}

	async load() {
		this.id = await tfrpc.rpc.databaseGet('id');
		this.ids = await tfrpc.rpc.getIdentities() ?? [];
		this.posts = JSON.parse(await tfrpc.rpc.databaseGet('posts') ?? '[]');
		this.published = await tfrpc.rpc.query(`
			SELECT author, timestamp, content FROM messages
			JOIN json_each(?) AS ids
			WHERE author = ids.value AND json_extract(content, '$.type') IN ('blog', 'blog-edit')
			ORDER BY timestamp DESC
		`, [JSON.stringify(this.ids)]);
	}

	id_change(event) {
		tfrpc.rpc.databaseSet('id', event.srcElement.value);
	}

	async add_post() {
		let next_id = parseInt(await tfrpc.rpc.databaseGet('next_id') || '0');
		let created = new Date().toISOString();
		this.posts = [].concat([{id: next_id, created: created, title: 'New Post'}], this.posts);
		await tfrpc.rpc.databaseSet('next_id', `${next_id + 1}`);
		await tfrpc.rpc.databaseSet(`post_${next_id}`, JSON.stringify({id: next_id, created: created}));
		await tfrpc.rpc.databaseSet('posts', JSON.stringify(this.posts));
	}

	async remove_post(index) {
		this.posts = [].concat(this.posts.slice(0, index), this.posts.slice(index + 1));
		await tfrpc.rpc.databaseSet('posts', JSON.stringify(this.posts));
		this.post = undefined	;
	}

	post_change(name, value) {
		let change = {};
		change[name] = value;
		change.modified = new Date();
		this.post = Object.assign({}, this.post, JSON.parse(JSON.stringify(change)));
		tfrpc.rpc.databaseSet(`post_${this.post.id}`, JSON.stringify(this.post));
		if (name == 'title') {
			console.log('updating title');
			for (let post of this.posts) {
				console.log(post.id, this.post);
				if (post.id == this.post.id) {
					console.log('found post to update');
					post.title = value;
				}
			}
			this.posts = [].concat([], this.posts);
			tfrpc.rpc.databaseSet(`posts`, JSON.stringify(this.posts));
		}
	}

	async load_post(id) {
		this.post = JSON.parse(await tfrpc.rpc.databaseGet(`post_${id}`) || '{}');
		this.post.id = id;
	}

	markdown(md) {
		let reader = new commonmark.Parser({safe: true});
		let writer = new commonmark.HtmlRenderer();
		let parsed = reader.parse(md || '');
		return writer.render(parsed);
	}

	async publish() {
		let body = Array.from(new TextEncoder().encode(this.post.body));
		let body_id = await tfrpc.rpc.store_blob(body);
		let message = {
			type: 'blog',
			title: this.post.title,
			summary: this.post.summary,
			blog: body_id,
			mentions: [
				{
					link: body_id,
					type: 'text/markdown',
					size: body.length,
				},
			],
		};
		console.log(message);
		tfrpc.rpc.appendMessage(this.id, message).then(function() {
			// TODO: Wish we had the message id here: this.post_change('published', [].concat([new Date().toISOString()], this.post.published ?? []) */
		}).catch(function(error) {
			alert(error.message);
		});
	}

	render_post() {
		if (!this.post) {
			return;
		}
		return html`
			<div style="display: flex; flex-direction: row">
				<div style="display: flex; flex-direction: column">
					<input type="checkbox" ?checked=${this.post.published} @input=${event =>  this.post_change('published', event.srcElement.checked)}></input>
					<input type="text" value=${this.post.title} @input=${event =>  this.post_change('title', event.srcElement.value)}></input>
					<textarea @input=${event =>  this.post_change('summary', event.srcElement.value)}>${this.post.summary}</textarea>
					<textarea @input=${event =>  this.post_change('body', event.srcElement.value)}>${this.post.body}</textarea>
					<input type="button" @click=${this.publish} value="Publish"></input>
				</div>
				<div style="display: flex; flex-direction: column">
					<div>${this.post.published ? 'published' : 'draft'}</div>
					<div>${unsafeHTML(this.markdown(this.post.title))}</div>
					<div>${unsafeHTML(this.markdown(this.post.summary))}</div>
					<div>${unsafeHTML(this.markdown(this.post.body))}</div>
					${Object.entries(this.post).filter(([k, v]) => ['title', 'summary', 'body'].indexOf(k) == -1).map(([k, v]) => html`<div>${k} = ${v}</div>`)}
				</div>
			</div>
		`;
	}

	render() {
		return html`
			<style>
				.post_item {
					cursor: pointer;
					display: flex;
					flex-direction: row;
					gap: 1em;
				}
				.post_item:hover {
					background-color: #ccc;
				}
			</style>
			<div><a href="./index">index</a><div>
			<select @change=${this.id_change}>
				${this.ids.map(id => html`<option ?selected=${this.id == id}>${id}</option>`)}
			</select>
			<ul>
				${this.posts.map((post, index) => html`<li class="post_item" style="${post.id == this.post?.id ? 'background-color: #888' : undefined}" @click=${() => this.load_post(post.id)}>
					<div><p>${post.created}</p></div>
					<div style="flex: 1 1">${unsafeHTML(this.markdown(post.title))}</div>
					<input type="button" value="x" @click=${event => this.remove_post(index)}></input>
				</li>`)}
			</ul>
			<pre>${JSON.stringify(this.published, null, 2)}</pre>
			<input type="button" @click=${this.add_post} value="Add Post"></input>
			${this.render_post()}
		`;
	}
}
customElements.define('blog-app', BlogAppElement);